// code by jph
package kah;

import java.util.Arrays;
import java.util.stream.IntStream;

import kah.var.Rules;
import sys.IntRange;
import sys.IntegerMath;

/**
 * Encodes state of board
 */
public class Board implements Cloneable {
  public static final int FOURTEEN = 14;
  public static final int KALAHA_USER = 6;
  public static final int KALAHA_SELF = 13;

  /**
   * @param seeds number of stones in each pit
   * @param rules of the game variant
   * @return board
   */
  public static Board init(int seeds, Rules rules) {
    Board myBoard = new Board(new int[FOURTEEN], rules);
    for (int index : new IntRange(KALAHA_USER)) {
      myBoard.val[index + 0] = seeds;
      myBoard.val[index + 7] = seeds;
    }
    return myBoard;
  }

  private int[] val;
  private final Rules rules;

  private Board(int[] val, Rules rules) {
    assert val.length == FOURTEEN;
    this.val = val;
    this.rules = rules;
  }

  /**
   * @param pos
   * @return number of stones in pit with index
   */
  public int get(int pos) {
    return val[pos];
  }

  /**
   * @return new instance of board "rotated by 180 degree"
   */
  public Board flip() {
    Board myBoard = new Board(new int[FOURTEEN], this.rules);
    for (int index : new IntRange(7)) {
      myBoard.val[index + 0] = val[index + 7];
      myBoard.val[index + 7] = val[index + 0];
    }
    return myBoard;
  }

  /**
   * @return new instance of board that is a copy of this board
   */
  public Board safe_clone() {
    return new Board(Arrays.copyOf(val, val.length), this.rules);
  }

  /**
   * @param pos
   * @return true if pit is non-empty
   */
  public boolean hasStones(int pos) {
    return 0 < val[pos];
  }

  /**
   * @param pos
   * @return true if pit with position is between 0, 1, ..., 5
   */
  private boolean isOwn(int pos) {
    return 0 <= pos && pos < KALAHA_USER;
  }

  /**
   * @param pos
   * @return index of pit that is opposite of given pit
   */
  private int opposite(int pos) {
    return 12 - pos;
  }

  /**
   * perform standard move:
   * stones are removed from pit and distributed consecutively into the adjacent pits 
   *  
   * @param pos position must index a non empty pit
   * @return new instance of board
   */
  public Board move(int pos) {
    return safe_clone().private_move(pos);
  }

  // helper function
  private Board private_move(int pos) {
    assert isOwn(pos);
    final int num = val[pos];
    assert 0 < num;
    // ---
    val[pos] = 0; // stones transfered to num
    int rep = -1; // address of increment
    for (int index : new IntRange(1, num + 1)) {
      rep = IntegerMath.mod(pos + rules.direction(index), KALAHA_SELF);
      ++val[rep];
    }
    int opp = opposite(rep);
    if (isOwn(rep) && val[rep] == 1 && //
        (val[opp] != 0 || rules.isEmptyCapture())) {
      val[6] += val[rep] + val[opp];
      val[rep] = 0;
      val[opp] = 0;
    }
    return this;
  }

  /**
   * @param pos
   * @return true, if selecting the pit at the given position
   * enables the same user to make another move right afterwards
   */
  public boolean isMoveAgain(int pos) {
    assert isOwn(pos);
    final int num = val[pos];
    assert 0 < num;
    return IntegerMath.mod(pos + rules.direction(num), KALAHA_SELF) == KALAHA_USER;
  }

  /**
   * @return true, if all pits of the user are empty
   */
  public boolean isMoveTerminal() {
    return !IntStream.of(val) //
        .limit(KALAHA_USER) //
        .filter(stones -> stones != 0) //
        .findAny() //
        .isPresent();
  }

  /**
   * transfer all stones from the pit at the opposite side
   * into the opponents Kahala.
   * 
   * move is valid only if the user doesn't have any stones in the pits.
   * 
   * @return new instance of board 
   */
  public Board moveTerminal() {
    assert isMoveTerminal();
    return safe_clone().private_moveTerminal();
  }

  // helper function
  private Board private_moveTerminal() {
    int sum = 0;
    for (int index : new IntRange(6)) {
      int pos = 7 + index;
      sum += val[pos];
      val[pos] = 0;
    }
    val[KALAHA_SELF] += rules.countRemaining(sum);
    return this;
  }

  // TODO function unused
  private boolean isEmpty() {
    return isMoveTerminal() && flip().isMoveTerminal();
  }

  // TODO function unused
  private boolean isDetermined() {
    int sum = IntStream.of(val).sum() / 2;
    return sum < val[KALAHA_SELF] || sum < val[KALAHA_USER];
  }

  /**
   * @return difference of stones in own Kalaha and opponents Kalaha
   */
  public int evaluate() {
    return val[KALAHA_USER] - val[KALAHA_SELF];
  }

  @Override
  public String toString() {
    StringBuilder myStringBuilder = new StringBuilder();
    for (int pos : new IntRange(6))
      myStringBuilder.append(String.format("%2d ", val[pos]));
    myStringBuilder.append(String.format(" (%2d) ", val[KALAHA_USER]));
    for (int pos : new IntRange(6))
      myStringBuilder.append(String.format("%2d ", val[7 + pos]));
    myStringBuilder.append(String.format(" (%2d) ", val[KALAHA_SELF]));
    return myStringBuilder.toString();
  }
}
