// code by jph
package kah;

import kah.mov.Move;
import kah.var.Rules;

/**
 * Kalah, also called Kalaha or Mancala,
 * is a game in the mancala family
 * invented in the United States by William Julius Champion, Jr
 * in 1940.
 * 
 * This game favors the starting player, 
 * who will always win the three-seed to 
 * six-seed versions with perfect play.
 */
public class Kalaha {
  private Board board;
  private boolean user;
  public Board board_display;
  public boolean user_display;

  public Kalaha(int stones, Rules rules, boolean user) {
    this(Board.init(stones, rules), user);
  }

  public Kalaha(Board board, boolean user) {
    this.board = board;
    this.user = user;
    board_display = board.safe_clone();
    user_display = user;
  }

  /**
   * @return true at turn of user, false at turn of computer 
   */
  public boolean isUser() {
    return user;
  }

  /**
   * function synchronizes display with internal game status
   * (purpose is to "slow down" computer moves)
   */
  public void advance() {
    board_display = board.safe_clone();
    user_display = user;
  }

  /**
   * call to function only permitted when isUser() returns true;
   * 
   * @param pos
   */
  public void userSelect(int pos) {
    assert user;
    if (user) {
      boolean again = false;
      if (board.isMoveTerminal()) {
        board = board.moveTerminal();
      } else {
        if (board.hasStones(pos)) {
          again = board.isMoveAgain(pos);
          board = board.move(pos);
        }
      }
      if (!again)
        user = false;
      advance(); // user moves advance immediately
    } else
      throw new RuntimeException(" !!! turn of self");
  }

  /**
   * function announces next computer move.
   * to finalize call advance() subsequently.
   * call to function only permitted when isUser() returns false;
   * 
   * @param depth
   * @return
   */
  public Minimax announceSelf(int flips) {
    assert !user;
    if (!user) {
      board_display = board.safe_clone();
      board = board.flip();
//      Minimax myMinimax = new DepthFirstMinimax(board, flips);
      Minimax myMinimax = new BreadthFirstMinimax(board, flips, 100000);
      Result myResult = myMinimax.myResult;
      if (myResult != null) {
        myMinimax.printInfo();
        boolean again = false;
        if (myResult.pos == Move.TERMINAL_KEY) {
          board = board.moveTerminal();
        } else {
          again = board.isMoveAgain(myResult.pos);
          board = board.move(myResult.pos);
        }
        if (!again)
          user = true;
      }
      board = board.flip();
      return myMinimax;
    }
    throw new RuntimeException(" !!! turn of user");
  }
}
